! Copyright 2020-2021 Max Planck Institute for Dynamics and Self-Organization
!
! Permission is hereby granted, free of charge, to any person obtaining
! a copy of this software and associated documentation files (the
! "Software"), to deal in the Software without restriction, including
! without limitation the rights to use, copy, modify, merge, publish,
! distribute, sublicense, and/or sell copies of the Software, and to
! permit persons to whom the Software is furnished to do so, subject to
! the following conditions:
!
! The above copyright notice and this permission notice shall be
! included in all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
! EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
! MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
! NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
! LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
! OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
! WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


    PURE LOGICAL FUNCTION ADDSUFFIX(is_vector_nonnan_nonnegative)(a, na)
        !! Returns whether all elements of the given vector are non-NaN and
        !! non-negative or not.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: na
        !! The number of elements in *a*.
        REAL(KIND=REALP()), DIMENSION(na), INTENT(IN) :: a
        !! The vector to validate.

        ! Iteration index over a.
        INTEGER(KIND=C_INT) :: ia

        ! Negative na is invalid. If na is non-negative, we must check the
        ! elements.
        IF (na < 0) THEN
            ADDSUFFIX(is_vector_nonnan_nonnegative) = .FALSE.
        ELSE
            DO ia = 1, na
                IF (IEEE_IS_NAN(a(ia)) .OR. a(ia) < REALPCONST(0.)) THEN
                    ADDSUFFIX(is_vector_nonnan_nonnegative) = .FALSE.
                    RETURN
                ENDIF
            END DO
            ADDSUFFIX(is_vector_nonnan_nonnegative) = .TRUE.
        ENDIF
    END FUNCTION ADDSUFFIX(is_vector_nonnan_nonnegative)


    PURE LOGICAL FUNCTION ADDSUFFIX(is_vector_zero_to_one)(a, na)
        !! Returns whether all elements of the given vector are non-NaN and
        !! between 0 and 1 inclusive or not.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: na
        !! The number of elements in *a*.
        REAL(KIND=REALP()), DIMENSION(na), INTENT(IN) :: a
        !! The vector to validate.

        ! Iteration index over a.
        INTEGER(KIND=C_INT) :: ia

        ! Negative na is invalid. If na is non-negative, we must check the
        ! elements.
        IF (na < 0) THEN
            ADDSUFFIX(is_vector_zero_to_one) = .FALSE.
        ELSE
            DO ia = 1, na
                IF ((.NOT. IEEE_IS_FINITE(a(ia))) &
                     & .OR. a(ia) < REALPCONST(0.) &
                     & .OR. a(ia) > REALPCONST(1.)) THEN
                    ADDSUFFIX(is_vector_zero_to_one) = .FALSE.
                    RETURN
                ENDIF
            END DO
            ADDSUFFIX(is_vector_zero_to_one) = .TRUE.
        ENDIF
    END FUNCTION ADDSUFFIX(is_vector_zero_to_one)


    PURE REAL(KIND=REALP()) FUNCTION ADDSUFFIX(required_emax_vks)( &
         & y, Mc, ny, validate_arrays) MAKECBIND
        !! Determine the approximate minimum value of the maximum supported
        !! exponent (emax) to prevent overflow when calculating the
        !! \(V_k\left(\vec{y},z\right)\). The final ceiling operation is not
        !! taken. Returns NaN if the arguments are invalid.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.

        ! We need variabiles to hold the maximum value we are considering from
        ! the array (maximum of 1 and the maximum element of y), Mc as a real,
        ! and the ceiling and floor of Mc / 2.
        REAL(KIND=REALP()) :: max_element, Mc_r, half_Mc_floor, half_Mc_ceil

        ! Handle the excpetional case of negative Mc or ny and return NaN. For
        ! Mc of 0 or ny of 0, we can just return 0.
        IF (Mc < 0 .OR. ny < 0) THEN
            ADDSUFFIX(required_emax_vks) = IEEE_VALUE(REALPCONST(0.), &
                 & IEEE_QUIET_NAN)
            RETURN
        ELSE IF (Mc == 0 .OR. ny == 0) THEN
            ADDSUFFIX(required_emax_vks) = REALPCONST(0.)
            RETURN
        ENDIF

        ! Optionally validate the elements of y.
        IF (validate_arrays .AND. .NOT. &
             & ADDSUFFIX(is_vector_nonnan_nonnegative)(y, ny * Mc)) THEN
            ADDSUFFIX(required_emax_vks) = &
                 & IEEE_VALUE(REALPCONST(0.), IEEE_QUIET_NAN)
            RETURN
        ENDIF

        ! For Mc of 1, it is just determined by the maximum element and the
        ! longer formula will give infinity due to the floor of Mc / 2.
        IF (Mc == 1) THEN
            ADDSUFFIX(required_emax_vks) = LOG(MAX(REALPCONST(1.), MAXVAL(y))) &
                 & / LOG(REALPCONST(2.))
            RETURN
        ENDIF

        ! Calculate them.
        max_element = MAX(REALPCONST(1.), MAXVAL(y))
        Mc_r = REAL(Mc, REALP())
        half_Mc_floor = REAL(FLOOR(Mc_r / REALPCONST(2.), C_INT), REALP())
        half_Mc_ceil = REAL(CEILING(Mc_r / REALPCONST(2.), C_INT), REALP())

        ADDSUFFIX(required_emax_vks) = (LOG(max_element) &
             & + (Mc_r + REALPCONST(1.5)) * LOG(Mc_r) &
             & - (half_Mc_floor + REALPCONST(0.5)) * LOG(half_Mc_floor) &
             & - (half_Mc_ceil + REALPCONST(0.5)) * LOG(half_Mc_ceil) &
             & + (REALPCONST(12.) * Mc_r)**(-1) &
             & - (REALPCONST(12.) * half_Mc_floor + REALPCONST(1.))**(-1) &
             & - (REALPCONST(12.) * half_Mc_ceil + REALPCONST(1.))**(-1) &
             & - REALPCONST(0.5) * LOG(REALPCONST(8.) * ATAN(REALPCONST(1.)))) &
             & / LOG(REALPCONST(2.))
    END FUNCTION ADDSUFFIX(required_emax_vks)


    PURE REAL(KIND=REALP()) FUNCTION ADDSUFFIX(required_emax_solution)( &
         & n0, beta, alpha, gamma, Mc, validate_arrays) MAKECBIND
        !! Determine the approximate minimum value of the maximum supported
        !! exponent (emax) to prevent overflow when calculating analytical
        !! recursive solution. The final ceiling operation is not taken.
        !! Returns NaN for invalid arguments.
        !!
        !! @warning
        !! One must add a safety factor to this value. This emax is gotten from
        !! considering the required emax to make sure
        !! \(\frac{1}{\gamma} V_k\left(\vec{y},z\right)\) does not overflow,
        !! but does not track the required exponent through the whole recursive
        !! formulas for the solution.
        !! @endwarning

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: n0
        !! The initial infectious aerosol concentration densities. Must be
        !! non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.

        ! We need variables to hold the individual required emaxes indicated
        ! for n0 and beta.
        REAL(KIND=REALP()) :: emax_n0, emax_beta

        ! For Mc, alpha, or gamma < 0 or NaN alpha or gamma, we need to
        ! indicate the arguments are bad. For Mc of zero, we can just return
        ! zero.
        IF (Mc < 0 .OR. IEEE_IS_NAN(alpha) .OR. alpha < REALPCONST(0.) &
             & .OR. IEEE_IS_NAN(gamma) .OR. gamma < REALPCONST(0.)) THEN
            ADDSUFFIX(required_emax_solution) = IEEE_VALUE(REALPCONST(0.), &
                 & IEEE_QUIET_NAN)
            RETURN
        ELSE IF (Mc == 0) THEN
            ADDSUFFIX(required_emax_solution) = REALPCONST(0.)
            RETURN
        ENDIF

        ! Optionally validate the elements of n0 and beta.
        IF (validate_arrays .AND. &
             & ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(n0, Mc)) &
             & .OR. &
             & (.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(beta, Mc)))) THEN
            ADDSUFFIX(required_emax_solution) = &
                 & IEEE_VALUE(REALPCONST(0.), IEEE_QUIET_NAN)
            RETURN
        ENDIF

        ! If gamma is 0, we don't have to worry about the recursive formula
        ! and can just consider the maximum of n0 and the maximum of beta
        ! divided by (alpha + gamma).
        !
        ! In both cases, we take choose the maximum emax of n0 and beta and
        ! scale it as if the other one had the same value (multiply by ln(2)).
        IF (gamma <= REALPCONST(0.)) THEN
            emax_n0 = LOG(MAXVAL(n0))
            emax_beta = LOG(MAXVAL(beta)) - LOG(alpha)
            ADDSUFFIX(required_emax_solution) = REALPCONST(1.) &
                 & + (MAX(emax_n0, emax_beta) / LOG(REALPCONST(2.)))
        ELSE
            emax_n0 = ADDSUFFIX(required_emax_vks)(n0, Mc, 1, .FALSE._C_BOOL)
            emax_beta = ADDSUFFIX(required_emax_vks)(beta, Mc, 1, .FALSE._C_BOOL)
            ADDSUFFIX(required_emax_solution) = REALPCONST(1.) &
                 & + MAX(emax_n0, emax_beta) &
                 & + (LOG(MAX(REALPCONST(1.), gamma**(-1))) &
                 & / LOG(REALPCONST(2.)))
        ENDIF
    END FUNCTION ADDSUFFIX(required_emax_solution)


    PURE INTEGER(KIND=C_INT) FUNCTION ADDSUFFIX(required_format_solution)( &
         & n0, beta, alpha, gamma, emax_safety_factor, Mc, &
         & validate_arrays) MAKECBIND
        !! Determine the floating point format with the smallest emax that
        !! shouldn't overflow when calculating the analytical solution.
        !!
        !! This function returns -1 if no format has a large enough emax, -2 if
        !! there is something wrong with the arguments, and positive otherwise
        !! with the following meanings.
        !!
        !! Value | Fortran Kind
        !! ----- | ---------------
        !! 1     | `C_FLOAT`
        !! 2     | `C_DOUBLE`
        !! 3     | `C_LONG_DOUBLE`
        !! 4     | `REAL128`
        !!
        !! @warning
        !! The safety factor *emax_safety_factor* is critical because the
        !! required emax is gotten from considering the required emax to make
        !! sure \(\frac{1}{\gamma} V_k\left(\vec{y},z\right)\) does not
        !! overflow, but does not track the required exponent through the whole
        !! recursive formulas for the solution. A sufficient safety factor
        !! helps prevent overflow when close to the limits of a particular
        !! floating point type.
        !! @endwarning

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: n0
        !! The initial infectious aerosol concentration densities. Must be
        !! non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        REAL(kind=REALP()), INTENT(IN), VALUE :: emax_safety_factor
        !! The safety factor to add to the estimated required emax. Must be
        !! non-negative.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.

        ! A variable is needed to hold the minimum required emax.
        REAL(kind=REALP()) :: required_emax

        ! The safety factor for emax must be non-NaN, finite, and non-negative.
        IF ((.NOT. IEEE_IS_FINITE(emax_safety_factor)) &
             & .OR. emax_safety_factor < REALPCONST(0.)) THEN
            ADDSUFFIX(required_format_solution) = FORMAT_VALUE_INVALID
            RETURN
        ENDIF

        ! Determine the required emax with the safety factor added. If it is
        ! NaN, an invalid argument was given.
        required_emax = ADDSUFFIX(required_emax_solution)( &
             & n0, beta, alpha, gamma, Mc, validate_arrays) + emax_safety_factor
        IF (IEEE_IS_NAN(required_emax)) THEN
            ADDSUFFIX(required_format_solution) = FORMAT_VALUE_INVALID
            RETURN
        ENDIF

        ! Find the available precision to use.
#ifdef LIBPMADRA_CFLOAT_SUPPORTED
        IF (required_emax <= REAL(MAXEXPONENT(0._C_FLOAT), REALP())) THEN
            ADDSUFFIX(required_format_solution) = FORMAT_VALUE_SP
            RETURN
        ENDIF
#endif
#ifdef LIBPMADRA_CDOUBLE_SUPPORTED
        IF (required_emax <= REAL(MAXEXPONENT(0._C_DOUBLE), REALP())) THEN
            ADDSUFFIX(required_format_solution) = FORMAT_VALUE_DP
            RETURN
        ENDIF
#endif
#ifdef LIBPMADRA_CLONGDOUBLE_SUPPORTED
        IF (required_emax <= REAL(MAXEXPONENT(0._C_LONG_DOUBLE), REALP())) THEN
            ADDSUFFIX(required_format_solution) = FORMAT_VALUE_LD
            RETURN
        ENDIF
#endif
#ifdef LIBPMADRA_REAL128_SUPPORTED
        IF (required_emax <= REAL(MAXEXPONENT(0._REAL128), REALP())) THEN
            ADDSUFFIX(required_format_solution) = FORMAT_VALUE_QP
            RETURN
        ENDIF
#endif

        ! None had a large enough emax.
        ADDSUFFIX(required_format_solution) = FORMAT_VALUE_NONE
    END FUNCTION ADDSUFFIX(required_format_solution)


    PURE INTEGER(KIND=C_INT) FUNCTION ADDSUFFIX(required_Mc_folded)( &
         & n0, beta, alpha, gamma, emax_safety_factor, Mc, desired_format, &
         & validate_arrays) MAKECBIND
        !! Determine the maximum value of Mc that *n0* and *beta* can be folded
        !! at and have the output of *required_format_solution_SUFFIX* indicate
        !! that the required floating point format is *desired_format*. This
        !! function returns this *Mc_folded* value, -1 if no value would
        !! satisfy this requirement (not even the value one), -2 if there is
        !! something wrong with the arguments, and -3 if there was an error
        !! allocating memory.
        !!
        !! @warning
        !! The safety factor *emax_safety_factor* is critical because the
        !! required emax is gotten from considering the required emax to make
        !! sure \(\frac{1}{\gamma} V_k\left(\vec{y},z\right)\) does not
        !! overflow, but does not track the required exponent through the whole
        !! recursive formulas for the solution. A sufficient safety factor
        !! helps prevent overflow when close to the limits of a particular
        !! floating point type.
        !! @endwarning

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: n0
        !! The initial infectious aerosol concentration densities. Must be
        !! non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: desired_format
        !! The largest desired floating point format, which must be one of the
        !! values in the following table along with their meanings.
        !!
        !! Value | Fortran Kind
        !! ----- | ---------------
        !! 1     | `C_FLOAT`
        !! 2     | `C_DOUBLE`
        !! 3     | `C_LONG_DOUBLE`
        !! 4     | `REAL128`
        REAL(kind=REALP()), INTENT(IN), VALUE :: emax_safety_factor
        !! The safety factor to add to the estimated required emax. Must be
        !! non-negative.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.

        ! We will find the proper Mc by testing different folding values and
        ! finding the value by a binary search. The upper bound will start as
        ! Mc. If the format is already small enough, then that is the folding
        ! value. Then, the initial lower bound of 1 is tested. If that is too
        ! big, then no folding will make a small enough format. Then, the
        ! binary search is done until Mc_upper == Mc_lower + 1 while keeping
        ! the format for Mc_lower no larger than desired_format and the format
        ! for Mc_upper being larger than the format.

        ! Need variables to hold the formats and Mc for the lower and upper
        ! bounds as well as the test values.
        INTEGER(KIND=C_INT) :: format_lower, format_upper, format_test
        INTEGER(KIND=C_INT) :: Mc_lower, Mc_upper, MC_test

        ! Arrays for the folded n0 and beta are needed, which will have to be
        ! allocated dynamically one time to have a length of Mc (don't need to
        ! reallocate to do smaller folding values).
        REAL(KIND=REALP()), DIMENSION(:), ALLOCATABLE :: n0_folded, beta_folded

        ! A status flag is needed for allocation and calls to fold_Mc, and a
        ! variable to hold the folded fraction (won't be used).
        INTEGER(KIND=C_INT) :: status
        REAL(KIND=REALP()), DIMENSION(1) :: folded_fraction

        ! desired_format must be between 1 and 4 inclusive.
        IF (desired_format < 1 .OR. desired_format > 4) THEN
            ADDSUFFIX(required_Mc_folded) = -2
            RETURN
        ENDIF

        ! Mc must be non-negative. Also, if Mc is zero, then no folding is
        ! necessary.
        IF (Mc < 0) THEN
            ADDSUFFIX(required_Mc_folded) = -2
            RETURN
        ELSE IF (Mc == 0) THEN
            ADDSUFFIX(required_Mc_folded) = 0
            RETURN
        ENDIF

        ! Get the required format without folding, which is the initial upper
        ! bound. This will also do all the required validation. If it is the
        ! desired format, we are done.
        format_upper = ADDSUFFIX(required_format_solution)(n0, beta, alpha, &
             & gamma, emax_safety_factor, Mc, validate_arrays)
        IF (format_upper == FORMAT_VALUE_INVALID) THEN
            ADDSUFFIX(required_Mc_folded) = -2
            RETURN
        ELSE IF (format_upper <= desired_format &
             & .AND. format_upper /= FORMAT_VALUE_NONE) THEN
            ADDSUFFIX(required_Mc_folded) = Mc
            RETURN
        ENDIF

        ! Allocate the folded arrays and return if something goes wrong.
        ALLOCATE(n0_folded(Mc), stat=status)
        IF (status /= 0) THEN
            ADDSUFFIX(required_Mc_folded) = -3
            RETURN
        ENDIF
        ALLOCATE(beta_folded(Mc), stat=status)
        IF (status /= 0) THEN
            ADDSUFFIX(required_Mc_folded) = -3
            RETURN
        ENDIF

        ! Get the required format for Mc_folded = 1, which is the initial lower
        ! bound. If it is too large, then no Mc_folded value will meet the requirements.
        CALL ADDSUFFIX(fold_Mc)(folded_fraction, n0_folded, n0, 1, Mc, 1, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) THEN
            ADDSUFFIX(required_Mc_folded) = status - 1
            RETURN
        ENDIF
        CALL ADDSUFFIX(fold_Mc)(folded_fraction, beta_folded, beta, 1, Mc, 1, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) THEN
            ADDSUFFIX(required_Mc_folded) = status - 1
            RETURN
        ENDIF
        format_lower = ADDSUFFIX(required_format_solution)(n0_folded, &
             & beta_folded, alpha, gamma, emax_safety_factor, 1, .FALSE._C_BOOL)
        IF (format_lower == FORMAT_VALUE_INVALID) THEN
            ADDSUFFIX(required_Mc_folded) = -2
            RETURN
        ELSE IF (format_lower > desired_format &
             & .OR. format_lower == FORMAT_VALUE_NONE) THEN
            ADDSUFFIX(required_Mc_folded) = -1
            RETURN
        ENDIF

        ! Do the binary search.
        Mc_upper = Mc
        Mc_lower = 1
        DO WHILE (Mc_upper > Mc_lower + 1)
            ! Truncating average of upper and lower Mc.
            Mc_test = (Mc_upper + Mc_lower) / 2
            ! Fold
            CALL ADDSUFFIX(fold_Mc)(folded_fraction, n0_folded, n0, Mc_test, &
                 & Mc, 1, .FALSE._C_BOOL, status)
            IF (status /= 0) THEN
                ADDSUFFIX(required_Mc_folded) = status - 1
                RETURN
            ENDIF
            CALL ADDSUFFIX(fold_Mc)(folded_fraction, beta_folded, beta, &
                 & Mc_test, Mc, 1, .FALSE._C_BOOL, status)
            IF (status /= 0) THEN
                ADDSUFFIX(required_Mc_folded) = status - 1
                RETURN
            ENDIF
            ! Test
            format_test = ADDSUFFIX(required_format_solution)(n0_folded, &
                 & beta_folded, alpha, gamma, emax_safety_factor, Mc_test, &
                 & .FALSE._C_BOOL)
            ! Check for invalid and adjust brackets.
            IF (format_test == FORMAT_VALUE_INVALID) THEN
                ADDSUFFIX(required_Mc_folded) = -2
                RETURN
            ELSE IF (format_test <= desired_format &
                 & .AND. format_test /= FORMAT_VALUE_NONE) THEN
                Mc_lower = Mc_test
                format_lower = format_test
            ELSE
                Mc_upper = Mc_test
                format_upper = format_test
            ENDIF
        END DO

        ! The lower bracket is the required value.
        ADDSUFFIX(required_Mc_folded) = Mc_lower

    END FUNCTION ADDSUFFIX(required_Mc_folded)


    PURE SUBROUTINE ADDSUFFIX(vk)(vk, y, x, Mc, ny, nx, k, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(V_k(\vec{y}, x)\) for one *k*, any number of y-vectors,
        !! and any number of x-values. The binomial coefficients are calculated
        !! with the help of the natural logarithm of the Gamma function.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered. Must be at least the value of
        !! *k*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(OUT) :: vk
        !! The array to put the Vk values in.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(nx), INTENT(IN) :: x
        !! The x-values. Must be between 0 and 1 inclusivie.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: k
        !! The multiplicity. Must be between 1 and *Mc* inclusive.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! Temporary variable to hold the natural logarithm of k! and the
        ! binomial coefficient itself.
        REAL(kind=REALP()) :: ln_k_factorial, bincoef

        ! The i in the equation, which is a looping index and a looping index
        ! for the x-values.
        INTEGER(KIND=C_INT) :: i, ix

        ! We require 1 <= k <= Mc and that Mc > 1 and ny and nx are
        ! non-negative. Otherwise, we have been given bad arguments.
        IF (k <= 0 .OR. k > Mc .OR. Mc <= 0 .OR. ny < 0 .OR. nx < 0) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x and y. Otherwise or if the
        ! testing passes, all is good.
        IF (validate_arrays .AND. &
             & ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) &
             & .OR. &
             & (.NOT. ADDSUFFIX(is_vector_zero_to_one)(x, nx)))) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is nothing to do if ny or nx are zero.
        IF (ny == 0 .OR. nx == 0) RETURN

        ! We only need to compute ln(k!) once.
        ln_k_factorial = LOG_GAMMA(REAL(k + 1, REALP()))

        ! We will initialize the vk all to y_k (i = k) and then go i-value by
        ! i-value starting from i = k + 1 and adding the result to them all
        ! (inner loops over the y-vectors and x-values) so that each binomial
        ! coefficient only has to be evaluated once.
        !
        ! For each i after evaluating the binomial coefficient, we loop over
        ! the x values (can be done in parralel) with the innermost loop over
        ! the y-vectors.
        DO CONCURRENT (ix = 1:nx)
            vk(:, ix) = y(:, k)
        END DO
        DO i = k + 1, Mc
            ! Calculate the binomial coefficient and calculate the vk.
            bincoef = EXP(LOG_GAMMA(REAL(i + 1, REALP())) &
                 & - LOG_GAMMA(REAL(i - k + 1, REALP())) - ln_k_factorial)
            DO CONCURRENT (ix = 1:nx)
                vk(ix, :) = vk(ix, :) &
                     & + (bincoef * (REALPCONST(1.) - x(ix))**(i - k)) * y(:, i)
            END DO
        END DO
    END SUBROUTINE ADDSUFFIX(vk)


    PURE SUBROUTINE ADDSUFFIX(vks)(vks, y, x, Mc, ny, nx, compute_by_k, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(V_k(\vec{y}, x)\) for all k, any number of y-vectors,
        !! and any number of x-values. The binomial coefficients are calculated
        !! with the help of the natural logarithm of the Gamma function.
        !!
        !! *compute_by_k* is a switch that controls the order in which the
        !! iteration takes place, which has consequences on the scaling of the
        !! performance.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered. Must be at least the value of
        !! *k*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny, Mc), INTENT(OUT) :: vks
        !! The array to put the Vk values in.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(nx), INTENT(IN) :: x
        !! The x-values. Must be between 0 and 1 inclusive.
        LOGICAL(C_BOOL), INTENT(IN), VALUE :: compute_by_k
        !! Whether the computation of the Vk should go one k at a time (if
        !! true), or if th iteration should be over i - k (if false). The
        !! latter is faster for a small number of y-vectors, but the former is
        !! faster for a large number of y-vectors because it scales better in
        !! y-vectors.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful, -1 if the arguments are invalid, and -2 if
        !! memory cannot be allocated.

        ! For both methods, we need the k in the equation, which is a
        ! looping index and a temporary variable to hold the natural logarithm
        ! of a factorial.
        INTEGER(KIND=C_INT) :: k
        REAL(kind=REALP()) :: temp_ln_factorial

        ! For either iteration method, we will need to store either (1 - x) or
        ! (1 - x)**(i - k) to save recomputation. This array will have to be
        ! allocated and have a status flag for the allocation.
        REAL(kind=REALP()), DIMENSION(:), ALLOCATABLE :: temp_one_minus_x_raised
        INTEGER(KIND=C_INT) :: allocation_status


        ! For iterating over i - k, we need an additional variable for i - k.
        INTEGER(KIND=C_INT) :: i_minus_k

        ! We require that Mc is positive and ny and nx are non-negative.
        ! Otherwise, we have bad arguments.
        IF (Mc <= 0 .OR. ny < 0 .OR. nx < 0) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x and y. Otherwise or if the
        ! testing passes, all is good.
        IF (validate_arrays .AND. &
             & ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) &
             & .OR. &
             & (.NOT. ADDSUFFIX(is_vector_zero_to_one)(x, nx)))) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is nothing to do if ny or nx are zero.
        IF (ny == 0 .OR. nx == 0) RETURN

        ! Allocate temp_one_minux_x and return if something goes wrong.
        ALLOCATE(temp_one_minus_x_raised(nx), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF

        ! The actual computation depends on what we are iterating over.
        IF (compute_by_k) THEN
            ! Doing the k one by one.
            !
            ! Each k
            DO k = 1, Mc
                ! We only need to compute ln(k!) once.
                temp_ln_factorial = LOG_GAMMA(REAL(k + 1, REALP()))

                BLOCK
                    ! We need the i in the equation (also a looping index), a
                    ! looping index for the y-vectors, and a variable to hold
                    ! the binomial coefficient.
                    INTEGER(KIND=C_INT) :: i, iy
                    REAL(KIND=REALP()) :: bincoef

                    ! We will initialize the vk all to y_k (i = k) and then go
                    ! i-value by i-value starting from i = k + 1 and adding the
                    ! result to them all (inner loops over the y-vectors and
                    ! x-values) so that each binomial coefficient only has to
                    ! be evaluated once.
                    !
                    ! For each i after evaluating the binomial coefficient, we
                    ! loop over the x values (can be done in parralel) with the
                    ! innermost loop over the y-vectors.
                    DO CONCURRENT (iy = 1:ny)
                        vks(:, iy, k) = y(iy, k)
                    END DO
                    DO i = k + 1, Mc
                        ! Calculate the binomial coefficient and calculate the
                        ! vk.
                        bincoef = EXP(LOG_GAMMA(REAL(i + 1, REALP())) &
                             & - LOG_GAMMA(REAL(i - k + 1, REALP())) &
                             & - temp_ln_factorial)
                        temp_one_minus_x_raised(:) = &
                             & bincoef * (REALPCONST(1.) - x)**(i - k)
                        DO CONCURRENT (iy = 1:ny)
                            vks(:, iy, k) = vks(:, iy, k) &
                                 & + temp_one_minus_x_raised * y(iy, i)
                        END DO
                    END DO
                END BLOCK
            END DO
        ELSE
            ! We will iterate over i - k. We can first just assign i - k = 0.
            BLOCK
                INTEGER(KIND=C_INT) :: iy
                DO CONCURRENT (k = 1:Mc, iy = 1:ny)
                    vks(:, iy, k) = y(iy, k)
                END DO
            END BLOCK

            ! Now for i - k > 0
            DO i_minus_k = 1, Mc - 1
                ! Calculate ln((i - k)!) and (1 - x)**(i - k) since they will be
                ! the same for the whole iteration.
                temp_ln_factorial = LOG_GAMMA(REAL(i_minus_k + 1, REALP()))
                temp_one_minus_x_raised(:) = (REALPCONST(1.) - x)**i_minus_k
                ! Iterate over each k until i reaches Mc.
                DO CONCURRENT (k = 1:(Mc - i_minus_k))
                    BLOCK
                        ! We need the i in the equation, a looping index for
                        ! the y-vectors, and a variable to hold the binomial
                        ! coefficient.
                        INTEGER(KIND=C_INT) :: i, iy
                        REAL(KIND=REALP()) :: bincoef
                        ! Calculate them and loop over the y-vectors.
                        i = k + i_minus_k
                        bincoef = EXP(LOG_GAMMA(REAL(i + 1, REALP())) &
                             & - LOG_GAMMA(REAL(k + 1, REALP())) &
                             & - temp_ln_factorial)
                        DO CONCURRENT (iy = 1:ny)
                            vks(:, iy, k) = vks(:, iy, k) &
                                 & + (bincoef * y(iy, i)) &
                                 & * temp_one_minus_x_raised
                        END DO
                    END BLOCK
                END DO
            END DO
        ENDIF
    END SUBROUTINE ADDSUFFIX(vks)


    PURE SUBROUTINE ADDSUFFIX(uk_top)(uk, y, alpha, gamma, Mc, ny, nx, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(U_{M_c}(\vec{y}, x)\) for any number of y-vectors
        !! and any number of x-values (result doesn't depend on them, so they
        !! don't have to be given as an argument). This is the top
        !! (\(k = M_c\)) value.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered. Must be at least the value of
        !! *k*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(OUT) :: uk
        !! The array to put the Uk values in.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! 1 / gamma * s
        REAL(KIND=REALP()) :: scale

        ! Loop index over the y-vectors.
        INTEGER(KIND=C_INT) ::iy

        ! We require that Mc >= 1, ny and nx are non-negative, and alpha and
        ! gamma are non-negative.
        IF (Mc <= 0 .OR. ny < 0 .OR. nx < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! alpha and gamma cannot be simultaneously zero.
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of y. Otherwise or if the testing
        ! passes, all is good.
        IF (validate_arrays .AND. .NOT. &
             & ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is nothing to do if ny is zero.
        IF (ny == 0) RETURN

        ! Calculate the values.
        scale = -REALPCONST(1.) / (alpha + REAL(Mc, REALP()) * gamma)
        DO CONCURRENT (iy = 1:ny)
            uk(:, iy) = scale * y(iy, Mc)
        END DO
    END SUBROUTINE ADDSUFFIX(uk_top)


    PURE SUBROUTINE ADDSUFFIX(uk_next)(uk, uk_before, vk, x, alpha, gamma, &
         & Mc, ny, nx, k, validate_arrays, status) MAKECBIND
        !! Calculate \(U_k(\vec{y}, x)\) for one *k* from the values for *k* + 1
        !! for any number of y-vectors (not actually passed since they aren't
        !! used) and any number of x-values. This subroutine does the next step
        !! of the recursion.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered. Must be greater than *k*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(OUT) :: uk
        !! The array to put the Uk values in.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(IN) :: uk_before
        !! The array of the Uk values for *k* + 1.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(IN) :: vk
        !! The Vk values for k.
        REAL(KIND=REALP()), DIMENSION(nx), INTENT(IN) :: x
        !! The x-values. Must be between 0 and 1 inclusive.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: k
        !! The multiplicity. Must be between 1 and *Mc* - 1 inclusive.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! The scalars multiplying U_{k+1} without the x and Vk in the solution.
        REAL(KIND=REALP()) :: scale_uk, scale_vk

        ! Loop index over the y-vectors.
        INTEGER(KIND=C_INT) :: iy

        ! We require that 1 <= k < Mc, Mc >= 1, ny and nx are non-negative,
        ! and alpha and gamma are non-negative.
        IF (k <= 0 .OR. k >= Mc .OR. Mc <= 0 .OR. ny < 0 .OR. nx < 0 &
             & .OR. IEEE_IS_NAN(alpha) .OR. alpha < REALPCONST(0.) &
             & .OR. IEEE_IS_NAN(gamma) .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! alpha and gamma cannot be simultaneously zero.
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x. Otherwise or if the testing
        ! passes, all is good.
        IF (validate_arrays .AND. .NOT. &
             & ADDSUFFIX(is_vector_zero_to_one)(x, nx)) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is nothing to do if ny or nx are zero.
        IF (ny == 0 .OR. nx == 0) RETURN

        ! Calculate the values.
        scale_uk = REAL(k + 1, REALP()) / ((alpha / gamma) + REAL(k, REALP()))
        scale_vk = REALPCONST(1.) / (alpha + REAL(k, REALP()) * gamma)
        DO CONCURRENT (iy = 1:ny)
            uk(:, iy) = scale_uk * x * uk_before(:, iy) - scale_vk * vk(:, iy)
        END DO
    END SUBROUTINE ADDSUFFIX(uk_next)


    PURE SUBROUTINE ADDSUFFIX(uk_one_next)(uk, uk_before, y, alpha, gamma, &
         & Mc, ny, k, validate_arrays, status) MAKECBIND
        !! Calculate \(U_k(\vec{y}, 1)\) for one *k* from the values for *k* + 1
        !! for any number of y-vectors. This subroutine does the next step of
        !! the recursion.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered. Must be greater than *k*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        REAL(KIND=REALP()), DIMENSION(ny), INTENT(OUT) :: uk
        !! The array to put the Uk values in.
        REAL(KIND=REALP()), DIMENSION(ny), INTENT(IN) :: uk_before
        !! The array of the Uk values for k + 1.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The x-values. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: k
        !! The multiplicity. Must be between 1 and *Mc* - 1 inclusive.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! We require that 1 <= k < Mc, Mc >= 1, ny and nx are non-negative, and
        ! alpha and gamma are non-negative.
        IF (k <= 0 .OR. k >= Mc .OR. Mc <= 0 .OR. ny < 0 &
             & .OR. IEEE_IS_NAN(alpha) .OR. alpha < REALPCONST(0.) &
             & .OR. IEEE_IS_NAN(gamma) .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! alpha and gamma cannot be simultaneously zero.
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of y. Otherwise or if the testing
        ! passes, all is good.
        IF (validate_arrays .AND. .NOT. &
             & ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is nothing to do if  ny is zero.
        IF (ny == 0) RETURN

        uk = (REAL(k + 1, REALP()) / &
             & ((alpha / gamma) + REAL(k, REALP()))) * uk_before &
             & - (REALPCONST(1.) / (alpha + REAL(k, REALP()) * gamma)) * y(:, k)
    END SUBROUTINE ADDSUFFIX(uk_one_next)


    PURE SUBROUTINE ADDSUFFIX(uks)(uks, vks, y, x, alpha, gamma, Mc, ny, nx, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(U_k(\vec{y}, x)\) for all k for any number of y-vectors
        !! and any number of x-values.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny, Mc), INTENT(OUT) :: uks
        !! The array to put the Uk values for each k in.
        REAL(KIND=REALP()), DIMENSION(nx, ny, Mc), INTENT(IN) :: vks
        !! The Vk values for each k.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(nx), INTENT(IN) :: x
        !! The x-values. Must be between 0 and 1 inclusive.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! The k in the equation, which is a looping index.
        INTEGER(KIND=C_INT) :: k

        ! We require that Mc is positive, ny and nx are non-negative, and alpha
        ! and gamma are non-negative. Otherwise, we have bad arguments.
        IF (Mc <= 0 .OR. ny < 0 .OR. nx < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! alpha and gamma cannot be simultaneously zero.
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x and y. Otherwise or if the
        ! testing passes, all is good.
        IF (validate_arrays .AND. &
             & ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) &
             & .OR. &
             & (.NOT. ADDSUFFIX(is_vector_zero_to_one)(x, nx)))) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is nothing to do if ny or nx are zero.
        IF (ny == 0 .OR. nx == 0) RETURN

        ! We have to start at the top of the recursion and work our way down.
        CALL ADDSUFFIX(uk_top)(uks(:, :, Mc), y, alpha, gamma, Mc, ny, nx, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN
        DO k = Mc - 1, 1, -1
            CALL ADDSUFFIX(uk_next)(uks(:, :, k), uks(:, :, k + 1), vks(:, :, k), x, &
                 & alpha, gamma, Mc, ny, nx, k, .FALSE._C_BOOL, status)
            IF (status /= 0) RETURN
        END DO
    END SUBROUTINE ADDSUFFIX(uks)


    PURE SUBROUTINE ADDSUFFIX(uks_one)(uks, y, alpha, gamma, Mc, ny, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(U_k(\vec{y}, 1)\) for all k for any number of y-vectors.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(OUT) :: uks
        !! The array to put the Uk values for each k in.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! The k in the equation, which is a looping index.
        INTEGER(KIND=C_INT) :: k

        ! We require that Mc is positive, ny is non-negative, and alpha and
        ! gamma are non-negative. Otherwise, we have bad arguments.
        IF (Mc <= 0 .OR. ny < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x and y. Otherwise or if the
        ! testing passes, all is good.
        IF (validate_arrays .AND. .NOT. &
             & ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is nothing to do if ny is zero.
        IF (ny == 0) RETURN

        ! We have to start at the top of the recursion and work our way down.
        CALL ADDSUFFIX(uk_top)(uks(:, Mc), y, alpha, gamma, Mc, ny, 1, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN
        DO k = Mc - 1, 1, -1
            CALL ADDSUFFIX(uk_one_next)(uks(:, k), uks(:, k + 1), y, &
                 & alpha, gamma, Mc, ny, k, .FALSE._C_BOOL, status)
            IF (status /= 0) RETURN
        END DO
    END SUBROUTINE ADDSUFFIX(uks_one)


    PURE SUBROUTINE ADDSUFFIX(wk_top)(wk, y, x, alpha, gamma, Mc, ny, nx, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(W_{M_c}(\vec{y}, x)\) for any number of y-vectors
        !! and any number of x-values. This is the top (\(k = M_c\)) value.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(OUT) :: wk
        !! The array to put the Wk values in.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(nx), INTENT(IN) :: x
        !! The x-values. Must be between 0 and 1 inclusive.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! The values for s and 1 / gamma * s**2
        REAL(KIND=REALP()) :: s, scale

        ! Loop indicex over the x-values
        INTEGER(KIND=C_INT) :: ix

        ! We require that Mc >= 1, ny and nx are non-negative, and alpha and
        ! gamma are non-negative. Otherwise, we have been given bad arguments.
        IF (Mc <= 0 .OR. ny < 0 .OR. nx < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! alpha and gamma cannot be simultaneously zero.
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x and y. Otherwise or if the
        ! testing passes, all is good.
        IF (validate_arrays .AND. &
             & ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) &
             & .OR. &
             & (.NOT. ADDSUFFIX(is_vector_zero_to_one)(x, nx)))) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is also nothing to do if ny or nx are zero.
        IF (ny == 0 .OR. nx == 0) RETURN

        ! Calculate the values.
        s = (alpha / gamma) + REAL(Mc, REALP())
        scale = REALPCONST(1.) / (s * (alpha + REAL(Mc, REALP()) * gamma))
        DO CONCURRENT (ix = 1:nx)
            wk(ix, :) = (scale * (REALPCONST(1.) - x(ix)**s)) * y(:, Mc)
        END DO
    END SUBROUTINE ADDSUFFIX(wk_top)


    PURE SUBROUTINE ADDSUFFIX(wk_next)(wk, wk_before, uk, uk_one, x, alpha, gamma, &
         & Mc, ny, nx, k, validate_arrays, status) MAKECBIND
        !! Calculate \(W_k(\vec{y}, x)\) for one *k* from the values for *k* + 1
        !! for any number of y-vectors (not passed since they aren't used) and
        !! any number of x-values. This subroutine does the next step in the
        !! recursion.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered. Must be greater than *k*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(OUT) :: wk
        !! The array to put the Wk values in.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(IN) :: wk_before
        !! The array of Wk values for k + 1.
        REAL(KIND=REALP()), DIMENSION(nx, ny), INTENT(IN) :: uk
        !! The array of Uk(y,x) values for k.
        REAL(KIND=REALP()), DIMENSION(ny), INTENT(IN) :: uk_one
        !! The array of Uk(y,1) values for k.
        REAL(KIND=REALP()), DIMENSION(nx), INTENT(IN) :: x
        !! The x-values. Must be between 0 and 1 inclusive.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: k
        !! The multiplicity. Must be between 1 and *Mc* - 1 inclusive.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! The values for s and 1 / s, and (k + 1) / s
        REAL(KIND=REALP()) :: s, inv_s, scale_wk

        ! Loop index over the x-values
        INTEGER(KIND=C_INT) :: ix

        ! We require that 1 <= k < Mc, Mc >= 1, ny and nx are non-negative, and
        ! alpha and gamma are non-negative.
        IF (k <= 0 .OR. k >= Mc .OR. Mc <= 0 .OR. ny < 0 .OR. nx < 0 &
             & .OR. IEEE_IS_NAN(alpha) .OR. alpha < REALPCONST(0.) &
             & .OR. IEEE_IS_NAN(gamma) .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! alpha and gamma cannot be simultaneously zero.
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x. Otherwise or if the testing
        ! passes, all is good.
        IF (validate_arrays .AND. .NOT. &
             & ADDSUFFIX(is_vector_zero_to_one)(x, nx)) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is also nothing to do if ny or nx are zero.
        IF (ny == 0 .OR. nx == 0) RETURN

        ! Calculate the values.
        s = (alpha / gamma) + REAL(k, REALP())
        inv_s = REALPCONST(1.) / s
        scale_wk = REAL(k + 1, REALP()) * inv_s
        DO CONCURRENT (ix = 1:nx)
            wk(ix, :) = scale_wk * wk_before(ix, :) &
                 & + inv_s * (x(ix)**s * uk(ix, :) &
                 & - uk_one)
        END DO
    END SUBROUTINE ADDSUFFIX(wk_next)


    PURE SUBROUTINE ADDSUFFIX(wks)(wks, uks, uks_one, y, x, alpha, gamma, Mc, ny, nx, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(W_k(\vec{y}, x)\) for all k for any number of y-vectors
        !! and any number of x-values.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: ny
        !! The number of y-vectors.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nx
        !! The number of x-values.
        REAL(KIND=REALP()), DIMENSION(nx, ny, Mc), INTENT(OUT) :: wks
        !! The array to put the Wk values for each k in.
        REAL(KIND=REALP()), DIMENSION(nx, ny, Mc), INTENT(IN) :: uks
        !! The Uk(y,x) values for each k.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: uks_one
        !! The Uk(y,1) values for each k.
        REAL(KIND=REALP()), DIMENSION(ny, Mc), INTENT(IN) :: y
        !! The y-vectors. Must be non-negative.
        REAL(KIND=REALP()), DIMENSION(nx), INTENT(IN) :: x
        !! The x-values. Must be between 0 and 1 inclusive.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative. Must not be simultaneously
        !! zero with *gamma*.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative. Must not be
        !! simultaneously zero with *alpha*.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! Loop index over the k.
        INTEGER(KIND=C_INT) :: k

        ! We require that Mc >= 1, ny and nx are non-negative, and alpha and
        ! gamma are non-negative. Otherwise, we have been given bad arguments.
        IF (Mc <= 0 .OR. ny < 0 .OR. nx < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! alpha and gamma cannot be simultaneously zero.
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of x and y. Otherwise or if the
        ! testing passes, all is good.
        IF (validate_arrays .AND. &
             & ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(y, Mc * ny)) &
             & .OR. &
             & (.NOT. ADDSUFFIX(is_vector_zero_to_one)(x, nx)))) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! There is also nothing to do if ny or nx are zero.
        IF (ny == 0 .OR. nx == 0) RETURN

        ! We have to start at the top of the recursion and work our way down.
        CALL ADDSUFFIX(wk_top)(wks(:, :, Mc), y, x, alpha, gamma, Mc, ny, nx, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN
        DO k = Mc - 1, 1, -1
            CALL ADDSUFFIX(wk_next)(wks(:, :, k), wks(:, :, k + 1), &
                 & uks(:, :, k), uks_one(:, k), &
                 & x, alpha, gamma, Mc, ny, nx, k, .FALSE._C_BOOL, status)
            IF (status /= 0) RETURN
        END DO
    END SUBROUTINE ADDSUFFIX(wks)


    PURE SUBROUTINE ADDSUFFIX(ninf)(ninf, beta, alpha, gamma, Mc, &
         & validate_arrays, status) MAKECBIND
        !! Calculate \(\vec{n}_\infty\) for all k, which is the infectious
        !! aerosol concentrations for each multiplicity as time goes to
        !! infinity (steady-state). The recursive formula is used.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(OUT) :: ninf
        !! The infectious aerosol concentration densities as time goes to
        !! infinity.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful and -1 if the arguments are invalid.

        ! We require that Mc is positive and beta, alpha, and gamma are
        ! non-negative. Otherwise, there are bad arguments.
        IF (Mc <= 0 .OR. IEEE_IS_NAN(alpha) .OR. alpha < REALPCONST(0.) &
             & .OR. IEEE_IS_NAN(gamma) .OR. gamma < REALPCONST(0.)) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of beta. Otherwise or if the testing
        ! passes, all is good.
        IF (validate_arrays .AND. .NOT. &
             & ADDSUFFIX(is_vector_nonnan_nonnegative)(beta, Mc)) THEN
            status = -1
            RETURN
        ENDIF
        status = 0

        ! We use the recursive formula, which even works for the special case
        ! of gamma = 0 unless alpha = 0 simultaneously. We have to handle this
        ! special case that both are zero (ninf is +inf for beta_k > 0 and 0
        ! for bet_k = 0).
        IF (alpha <= REALPCONST(0.) .AND. gamma <= REALPCONST(0.)) THEN
            WHERE (beta > REALPCONST(0.))
                ninf = IEEE_VALUE(REALPCONST(0.), IEEE_POSITIVE_INF)
            ELSEWHERE
                ninf = REALPCONST(0.)
            END WHERE
        ELSE
            CALL ADDSUFFIX(uks_one)(ninf, beta, alpha, gamma, Mc, 1, &
                 & .FALSE._C_BOOL, status)
            IF (status == 0) ninf = -ninf
        ENDIF
    END SUBROUTINE ADDSUFFIX(ninf)


    PURE SUBROUTINE ADDSUFFIX(solve_analytical_ninf_n)(ninf, n, n0, t, &
         & beta, alpha, gamma, t0, Mc, nt, validate_arrays, status) MAKECBIND
        !! Calculate the infectious aerosol concentrations for all k as time
        !! goes to infinity \(\vec{n}_\infty\) and at the provided times
        !! \(\vec{n}\). The recursive formulas are used.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nt
        !! The number of times.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(OUT) :: ninf
        !! The infectious aerosol concentration densities as time goes to
        !! infinity.
        REAL(KIND=REALP()), DIMENSION(nt, Mc), INTENT(OUT) :: n
        !! The infectious aerosol concentration densities at each time.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: n0
        !! The initial infectious aerosol concentration densities at time *t0*.
        !! Must be non-negative.
        REAL(kind=REALP()), DIMENSION(nt), INTENT(IN) :: t
        !! The times to find the infectious concentration densities at. Must be
        !! greater than or equal to *t0*.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        REAL(kind=REALP()), INTENT(IN), VALUE :: t0
        !! The initial time. Must be less than or equal to every element of
        !! *t*. Must be non-NaN and finite.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful, -1 if the arguments are invalid, and -2 if
        !! memory cannot be allocated.

        ! The k, which a loop index; a looping index for time; and a variable
        ! to hold the alpha / gamma (sbase)
        INTEGER(KIND=C_INT) :: k, it
        REAL(kind=REALP()) :: sbase

        ! We need to calculate z to pass to other proceedures , z**s, and also
        ! calculate both Vk(beta,z) and Vk(n0,z) togeter using the faster
        ! method (bulk calculation), so we will need to allocate arrays to hold
        ! them all. We also need arrays to hold the combined beta and n0
        ! vectors and the Uk(beta, z) for k and k + 1. We also need a variable
        ! to hold the allocation status.
        REAL(kind=REALP()), DIMENSION(:), ALLOCATABLE :: z, z_to_s
        REAL(kind=REALP()), DIMENSION(:, :), ALLOCATABLE :: beta_n0
        REAL(kind=REALP()), DIMENSION(:, :, :), ALLOCATABLE :: vks_beta_n0
        REAL(kind=REALP()), DIMENSION(:), ALLOCATABLE :: temp_uk_beta, &
             & temp_uk_beta_before
        INTEGER(KIND=C_INT) :: allocation_status

        ! We require that Mc is positive; nt, alpha, and gamma are
        ! non-negative; and t0 is non-NaN and finite. Otherwise, there are bad
        ! arguments.
        IF (Mc <= 0 .OR. nt < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.) &
             & .OR. (.NOT. IEEE_IS_FINITE(t0))) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of n0, beta, and t.
        IF (validate_arrays) THEN
            IF ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(n0, Mc)) &
                 & .OR. (.NOT. &
                 & ADDSUFFIX(is_vector_nonnan_nonnegative)(beta, Mc))) THEN
                status = -1
                RETURN
            ENDIF
            DO it = 1, nt
                IF (IEEE_IS_NAN(t(it)) .OR. t(it) < t0) THEN
                    status = -1
                    RETURN
                ENDIF
            END DO
        ENDIF

        ! Get ninf. Afterwards, if something went wrong in the calculation of
        ! ninf or if nt is zero, we are done.
        CALL ADDSUFFIX(ninf)(ninf, beta, alpha, gamma, Mc, .FALSE._C_BOOL, status)
        IF (status /= 0 .OR. nt == 0) RETURN

        ! Allocate z, returning and informing the caller if we are unable,
        ! which is the only one we need if gamma is zero.
        ALLOCATE(z(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF

        ! If gamma is zero, the result is trivial and can be easily calculated
        ! and the recursive formula doesn't have to be used.
        IF (gamma <= REALPCONST(0.)) THEN
            ! If alpha = 0, then the solution is linear:
            !
            !   n = n0 + beta * (t - t0).
            !
            ! Otherwise, it is
            !
            !   n = ninf + (n0 - ninf) * e^{-alpha (t - t0)}.
            !
            ! We will use z to hold t - t0 and e^{-alpha (t - t0)}
            ! respecitvely, instead of e^{-gamma (t - t0)}
            IF (alpha <= REALPCONST(0.)) THEN
                z(:) = t - t0
                DO CONCURRENT (k = 1:Mc)
                    n(:, k) = n0(k) + beta(k) * z
                END DO
            ELSE
                z(:) = EXP(-alpha * (t - t0))
                DO CONCURRENT (k = 1:Mc)
                    n(:, k) = z * n0(k) + (REALPCONST(1.) - z) * ninf(k)
                END DO
            ENDIF
            RETURN
        ENDIF

        ! Allocate the other temporary arrays, returning and informing the
        ! caller if we are unable.
        ALLOCATE(z_to_s(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(beta_n0(2, Mc), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(vks_beta_n0(nt, 2, Mc), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(temp_uk_beta(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(temp_uk_beta_before(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF

        ! Calculate z and sbase.
        z(:) = EXP(-gamma * (t - t0))
        sbase = alpha / gamma

        ! Calculate each n_k going backwards from the top. For each one, we
        ! first calculate U_k(beta, z) into n_k, then add V_k(n0, z), multiply
        ! by z**s, and then add n_inf,k.
        !
        ! Build beta_n0 and calculate all the vks.
        beta_n0(1, :) = beta
        beta_n0(2, :) = n0
        CALL ADDSUFFIX(vks)(vks_beta_n0, beta_n0, z, Mc, 2, nt, &
             & .FALSE._C_BOOL, .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN

        ! Handle the top (k = Mc).
        CALL ADDSUFFIX(uk_top)(temp_uk_beta, beta, alpha, gamma, Mc, 1, nt, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN
        z_to_s(:) = z**(sbase + REAL(Mc, REALP()))
        n(:, Mc) = z_to_s * vks_beta_n0(:, 2, Mc) &
             & + (ninf(Mc) + z_to_s * temp_uk_beta)

        ! Handle each successively lower k.
        DO k = Mc - 1, 1, -1
            ! Copy the temporary uk into the one for before, and then calculate
            ! the next! uk.
            temp_uk_beta_before(:) = temp_uk_beta
            CALL ADDSUFFIX(uk_next)(temp_uk_beta, temp_uk_beta_before, &
                 & vks_beta_n0(:, 1, k), z, alpha, gamma, Mc, 1, nt, k, &
                 & .FALSE._C_BOOL, status)
            IF (status /= 0) RETURN
            z_to_s(:) = z**(sbase + REAL(k, REALP()))
            n(:, k) = z_to_s * vks_beta_n0(:, 2, k) &
                 & + (ninf(k) + z_to_s * temp_uk_beta)
        END DO
    END SUBROUTINE ADDSUFFIX(solve_analytical_ninf_n)


    PURE SUBROUTINE ADDSUFFIX(solve_analytical_ninf_n_intn)(ninf, n, intn, n0, intn0, &
         & t, beta, alpha, gamma, t0, Mc, nt, validate_arrays, status) MAKECBIND
        !! Calculate the infectious aerosol concentrations for all k as time
        !! goes to infinity \(\vec{n}_\infty\) and at the provided times
        !! \(\vec{n}\) as well as its time integral at the provided times. The
        !! recursive formulas are used.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nt
        !! The number of times.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(OUT) :: ninf
        !! The infectious aerosol concentration densities as time goes to
        !! infinity.
        REAL(KIND=REALP()), DIMENSION(nt, Mc), INTENT(OUT) :: n
        !! The infectious aerosol concentration densities at each time.
        REAL(KIND=REALP()), DIMENSION(nt, Mc), INTENT(OUT) :: intn
        !! The time integrals of *n* at *t*.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: n0
        !! The initial infectious aerosol concentration densities at time *t0*.
        !! Must be non-negative.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: intn0
        !! The initial time integral of *n* values at *t0*. Must be
        !! non-negative.
        REAL(kind=REALP()), DIMENSION(nt), INTENT(IN) :: t
        !! The times to find the infectious concentration densities at. Must be
        !! greater than or equal to *t0*.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        REAL(kind=REALP()), INTENT(IN), VALUE :: t0
        !! The initial time. Must be less than or equal to every element of
        !! *t*. Must be non-NaN and finite.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful, -1 if the arguments are invalid, and -2 if
        !! memory cannot be allocated.

        ! The k, which a loop index; a looping index for time; a variable
        ! to hold the alpha / gamma (sbase); and a variable to hold the
        ! multiplicative inverse of alpha or gamma for scaling.
        INTEGER(KIND=C_INT) :: k, it
        REAL(kind=REALP()) :: sbase, scale

        ! We need to calculate t - t0, z to pass to other proceedures, and also
        ! calculate both Vk(beta,z) and Vk(n0,z) togeter using the faster
        ! method (bulk calculation), so we will need to allocate arrays to hold
        ! them all. We also need arrays to hold the combined beta and n0
        ! vectors, the Uk at x = 1, Uk of beta and n0 at k and k + 1, Wk
        ! of beta at k and k + 1, and z**s. We also need a variable to hold the
        ! allocation status.
        REAL(kind=REALP()), DIMENSION(:), ALLOCATABLE :: delta_t, z, z_to_s
        REAL(kind=REALP()), DIMENSION(:, :), ALLOCATABLE :: beta_n0
        REAL(kind=REALP()), DIMENSION(:, :, :), ALLOCATABLE :: vks_beta_n0
        REAL(kind=REALP()), DIMENSION(:, :), ALLOCATABLE :: uks_beta_n0_one
        REAL(kind=REALP()), DIMENSION(:, :), ALLOCATABLE :: uk_beta_n0, &
             & uk_beta_n0_before
        REAL(kind=REALP()), DIMENSION(:), ALLOCATABLE :: wk_beta, wk_beta_before
        INTEGER(KIND=C_INT) :: allocation_status

        ! We require that Mc is positive; nt, alpha, and gamma are
        ! non-negative; and t0 is non-NaN and finite. Otherwise, there are bad
        ! arguments.
        IF (Mc <= 0 .OR. nt < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.) &
             & .OR. (.NOT. IEEE_IS_FINITE(t0))) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of n0, intn0, beta, and t.
        IF (validate_arrays) THEN
            IF ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(n0, Mc)) &
                 & .OR. (.NOT. &
                 & ADDSUFFIX(is_vector_nonnan_nonnegative)(intn0, Mc)) &
                 & .OR. (.NOT. &
                 & ADDSUFFIX(is_vector_nonnan_nonnegative)(beta, Mc))) THEN
                status = -1
                RETURN
            ENDIF
            DO it = 1, nt
                IF (IEEE_IS_NAN(t(it)) .OR. t(it) < t0) THEN
                    status = -1
                    RETURN
                ENDIF
            END DO
        ENDIF

        ! Allocate the time related arrays, returning and informing the caller
        ! if we are unable.
        ALLOCATE(delta_t(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(z(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(z_to_s(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF

        ! Set delta_t
        delta_t(:) = t - t0

        ! If gamma is zero, the result is trivial and can be easily calculated
        ! and the recursive formula doesn't have to be used except for ninf.
        IF (gamma <= REALPCONST(0.)) THEN
            ! Get ninf. Afterwards, if something went wrong in the calculation
            ! of ninf or if nt is zero, we are done.
            CALL ADDSUFFIX(ninf)(ninf, beta, alpha, gamma, Mc, &
                 & .FALSE._C_BOOL, status)
            IF (status /= 0 .OR. nt == 0) RETURN
            ! If alpha = 0, then the solution is linear:
            !
            !   n = n0 + beta * (t - t0).
            !   intn = intn0 + n0 * (t - t0) + (beta/2) * (t - t0)**2
            !
            ! Otherwise, it is
            !
            !   n = ninf + (n0 - ninf) * e^{-alpha (t - t0)}.
            !   intn = intn0 + ninf * (t - t0)
            !          + (1/alpha) * (n0 - ninf) * (1 - e^{-alpha (t - t0)})
            !
            ! We will use z to hold (t - t0)**2 / 2 and e^{-alpha (t - t0)}
            ! respectively, instead of e^{-GAMMA (t - t0)}. We will use z_to_s
            ! to hold 1 - z (the new z).
            IF (alpha <= REALPCONST(0.)) THEN
                z(:) = REALPCONST(0.5) * delta_t**2
                DO CONCURRENT (k = 1:Mc)
                    n(:, k) = n0(k) + beta(k) * delta_t
                    intn(:, k) = intn0(k) + n0(k) * delta_t + beta(k) * z
                END DO
            ELSE
                z(:) = EXP(-alpha * (t - t0))
                z_to_s(:) = REALPCONST(1.) - z
                scale = REALPCONST(1.) / alpha
                DO CONCURRENT (k = 1:Mc)
                    n(:, k) = z * n0(k) + (REALPCONST(1.) - z) * ninf(k)
                    intn(:, k) = intn0(k) + ninf(k) * delta_t &
                         & + scale * (n0(k) - ninf(k)) * z_to_s
                END DO
            ENDIF
            RETURN
        ENDIF

        ! If there are no times, we just need to calculate ninf and we are
        ! done.
        IF (nt == 0) THEN
            CALL ADDSUFFIX(ninf)(ninf, beta, alpha, gamma, Mc, &
                 & .FALSE._C_BOOL, status)
            RETURN
        ENDIF

        ! Allocate the other temporary arrays, returning and informing the
        ! caller if we are unable.
        ALLOCATE(beta_n0(2, Mc), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(vks_beta_n0(nt, 2, Mc), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(uks_beta_n0_one(2, Mc), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(uk_beta_n0(nt, 2), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(uk_beta_n0_before(nt, 2), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(wk_beta(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF
        ALLOCATE(wk_beta_before(nt), stat=allocation_status)
        IF (allocation_status /= 0) THEN
            status = -2
            RETURN
        ENDIF

        ! Build beta_n0.
        beta_n0(1, :) = beta
        beta_n0(2, :) = n0

        ! Calculate uk_beta_n0_one and extract ninf.
        CALL ADDSUFFIX(uks_one)(uks_beta_n0_one, beta_n0, alpha, gamma, Mc, 2, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN
        ninf = -uks_beta_n0_one(1, :)

        ! Calculate z, sbase, and scale.
        z(:) = EXP(-gamma * delta_t)
        scale = REALPCONST(1.) / gamma
        sbase = scale * alpha

        ! Calculate each n_k going backwards from the top. For each one, we
        ! first calculate U_k(beta, z) into n_k, then add V_k(n0, z), multiply
        ! by z**s, and then add n_inf,k.
        !
        ! Calculate all the vks.
        CALL ADDSUFFIX(vks)(vks_beta_n0, beta_n0, z, Mc, 2, nt, &
             & .FALSE._C_BOOL, .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN

        ! Handle the top (k = Mc).
        CALL ADDSUFFIX(uk_top)(uk_beta_n0, beta_n0, alpha, gamma, Mc, 2, nt, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN
        CALL ADDSUFFIX(wk_top)(wk_beta, beta, z, alpha, gamma, Mc, 1, nt, &
             & .FALSE._C_BOOL, status)
        IF (status /= 0) RETURN
        z_to_s(:) = z**(sbase + REAL(Mc, REALP()))
        n(:, Mc) = z_to_s * vks_beta_n0(:, 2, Mc) &
             & + (ninf(Mc) + z_to_s * uk_beta_n0(:, 1))
        intn(:, Mc) = intn0(Mc) + ninf(Mc) * delta_t &
             & + (z_to_s * uk_beta_n0(:, 2) - uks_beta_n0_one(2, Mc))  &
             & - scale * wk_beta

        ! Handle each successively lower k.
        DO k = Mc - 1, 1, -1
            ! Calculate z**s.
            z_to_s(:) = z**(sbase + REAL(k, REALP()))
            ! Calculate the new uk from the previous.
            uk_beta_n0_before(:, :) = uk_beta_n0
            CALL ADDSUFFIX(uk_next)(uk_beta_n0, uk_beta_n0_before, &
                 & vks_beta_n0(:, :, k), z, alpha, gamma, Mc, 2, nt, k, &
                 .FALSE._C_BOOL, status)
            IF (status /= 0) RETURN
            ! Calculate the new wk from the previous.
            wk_beta_before(:) = wk_beta
            CALL ADDSUFFIX(wk_next)(wk_beta, wk_beta_before, uk_beta_n0(:, 1), &
                 & uks_beta_n0_one(1, k), z, alpha, gamma, Mc, 1, nt, k, &
                 & .FALSE._C_BOOL, status)
            IF (status /= 0) RETURN
            ! Calculate n and intn
            n(:, k) = z_to_s * vks_beta_n0(:, 2, k) &
                 & + (ninf(k) + z_to_s * uk_beta_n0(:, 1))
            intn(:, k) = intn0(k) + ninf(k) * delta_t &
                 & + (z_to_s * uk_beta_n0(:, 2) - uks_beta_n0_one(2, k))  &
                 & - scale * wk_beta
        END DO
    END SUBROUTINE ADDSUFFIX(solve_analytical_ninf_n_intn)


    PURE SUBROUTINE ADDSUFFIX(auto_solve_analytical_ninf_n)(ninf, n, n0, t, &
         & beta, alpha, gamma, emax_safety_factor, t0, Mc, nt, use_format, &
         & validate_arrays, format_used, status) MAKECBIND
        !! Calculate the infectious aerosol concentrations for all k as time
        !! goes to infinity \(\vec{n}_\infty\) and at the provided times
        !! \(\vec{n}\) in the specified floating point format, or the smallest
        !! available floating point precision that can do the computation
        !! without overflowing if the format should be determined automatically
        !! (see explanation of *use_format*). The recursive formulas are used.
        !! If the floating point type with the largest emax is insufficient
        !! when selecting the format automatically, the computation is not done
        !! and the situation is flagged.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nt
        !! The number of times.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(OUT) :: ninf
        !! The infectious aerosol concentration densities as time goes to
        !! infinity.
        REAL(KIND=REALP()), DIMENSION(nt, Mc), INTENT(OUT) :: n
        !! The infectious aerosol concentration densities at each time.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: n0
        !! The initial infectious aerosol concentration densities at time *t0*.
        !! Must be non-negative.
        REAL(kind=REALP()), DIMENSION(nt), INTENT(IN) :: t
        !! The times to find the infectious concentration densities at. Must be
        !! greater than or equal to *t0*
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        REAL(kind=REALP()), INTENT(IN), VALUE :: t0
        !! The initial time. Must be less than or equal to every element of
        !! *t*. Must be non-NaN and finite.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: use_format
        !! The floating point format that should be used, or 0 to determine it
        !! automatically. See *format_used* for the values for the different
        !! floating point formats.
        REAL(kind=REALP()), INTENT(IN), VALUE :: emax_safety_factor
        !! The safety factor to add to the estimated required emax. Must be
        !! non-negative.
        INTEGER(KIND=C_INT), INTENT(OUT) :: format_used
        !! The floating point format that actually ended up being used. This is
        !! -1 if none were determined to have a large enough emax and positive
        !! otherwise, with the following meanings.
        !!
        !! Value | Fortran Kind
        !! ----- | ---------------
        !! 1     | `C_FLOAT`
        !! 2     | `C_DOUBLE`
        !! 3     | `C_LONG_DOUBLE`
        !! 4     | `REAL128`
        !!
        !! If there is an error, its value is undefined.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful, -1 if the arguments are invalid, -2 if
        !! memory cannot be allocated, and -3 if no available floating point
        !! format has a large enough emax.

        ! Looping index over t for validation.
        INTEGER(KIND=C_INT) :: it

        ! A variable is required to hold allocation status.
        INTEGER :: allocation_status

        ! Variables are required to hold the converted values of the arguments
        ! to be able to pass to the subroutine for the right precision.
#if defined(LIBPMADRA_CFLOAT_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_SP
        REAL(KIND=C_FLOAT), DIMENSION(:), ALLOCATABLE :: ninf_sp, n0_sp, &
             & beta_sp, t_sp
        REAL(KIND=C_FLOAT), DIMENSION(:, :), ALLOCATABLE :: n_sp
        REAL(KIND=C_FLOAT) :: alpha_sp, gamma_sp, t0_sp
#endif
#if defined(LIBPMADRA_CDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_DP
        REAL(KIND=C_DOUBLE), DIMENSION(:), ALLOCATABLE :: ninf_dp, n0_dp, &
             & beta_dp, t_dp
        REAL(KIND=C_DOUBLE), DIMENSION(:, :), ALLOCATABLE :: n_dp
        REAL(KIND=C_DOUBLE) :: alpha_dp, gamma_dp, t0_dp
#endif
#if defined(LIBPMADRA_CLONGDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_LD
        REAL(KIND=C_LONG_DOUBLE), DIMENSION(:), ALLOCATABLE :: ninf_ld, n0_ld, &
             & beta_ld, t_ld
        REAL(KIND=C_LONG_DOUBLE), DIMENSION(:, :), ALLOCATABLE :: n_ld
        REAL(KIND=C_LONG_DOUBLE) :: alpha_ld, gamma_ld, t0_ld
#endif
#if defined(LIBPMADRA_REAL128_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_QP
        REAL(KIND=REAL128), DIMENSION(:), ALLOCATABLE :: ninf_qp, n0_qp, &
             & beta_qp, t_qp
        REAL(KIND=REAL128), DIMENSION(:, :), ALLOCATABLE :: n_qp
        REAL(KIND=REAL128) :: alpha_qp, gamma_qp, t0_qp
#endif

        ! We require that Mc is positive; nt, alpha, and gamma are
        ! non-negative; and t0 is non-NaN and finite. Otherwise, there are bad
        ! arguments.
        IF (Mc <= 0 .OR. nt < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.) &
             & .OR. (.NOT. IEEE_IS_FINITE(t0))) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of n0, beta, and t.
        IF (validate_arrays) THEN
            IF ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(n0, Mc)) &
                 & .OR. (.NOT. &
                 & ADDSUFFIX(is_vector_nonnan_nonnegative)(beta, Mc))) THEN
                status = -1
                RETURN
            ENDIF
            DO it = 1, nt
                IF (IEEE_IS_NAN(t(it)) .OR. t(it) < t0) THEN
                    status = -1
                    RETURN
                ENDIF
            END DO
        ENDIF

        ! Check use_format.
        IF (use_format < 0 .OR. use_format > 4) THEN
            status = -1
            format_used = FORMAT_VALUE_INVALID
            RETURN
        ENDIF

        ! Determine the format that should be used. If it is -2, then we need
        ! to stop.
        IF (use_format /= 0) THEN
            format_used = use_format
        ELSE
            format_used = ADDSUFFIX(required_format_solution)(n0, beta, &
                 & alpha, gamma, emax_safety_factor, Mc, .FALSE._C_BOOL)
            IF (format_used == FORMAT_VALUE_NONE) THEN
                status = -3
                RETURN
            ELSE IF (format_used < 1) THEN
                status = -1
                RETURN
            ENDIF
        ENDIF

        ! If the required format is the same as this procedure, we can just
        ! call it.
        IF (format_used == ACTUAL_FORMAT) THEN
            CALL ADDSUFFIX(solve_analytical_ninf_n)(ninf, n, n0, t, beta, &
                 & alpha, gamma, t0, Mc, nt, .FALSE._C_BOOL, status)
            RETURN
        ENDIF

        ! Now, we must convert the input arguments to the right type, allocate
        ! new outputs in the right type, call the right subroutine, and convert
        ! the new outputs to the right type.

        ! C_FLOAT
#if defined(LIBPMADRA_CFLOAT_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_SP
#define NEWSUFFIX(s) ADDSUFFIX_SP(s)
#define FORMAT_VALUE FORMAT_VALUE_SP
#define NEWREALP() C_FLOAT
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

        ! C_DOUBLE
#if defined(LIBPMADRA_CDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_DP
#define NEWSUFFIX(s) ADDSUFFIX_DP(s)
#define FORMAT_VALUE FORMAT_VALUE_DP
#define NEWREALP() C_DOUBLE
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

        ! C_LONG_DOUBLE
#if defined(LIBPMADRA_CLONGDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_LD
#define NEWSUFFIX(s) ADDSUFFIX_LD(s)
#define FORMAT_VALUE FORMAT_VALUE_LD
#define NEWREALP() C_LONG_DOUBLE
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

        ! REAL128
#if defined(LIBPMADRA_REAL128_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_QP
#define NEWSUFFIX(s) ADDSUFFIX_QP(s)
#define FORMAT_VALUE FORMAT_VALUE_QP
#define NEWREALP() REAL128
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

    END SUBROUTINE ADDSUFFIX(auto_solve_analytical_ninf_n)


    PURE SUBROUTINE ADDSUFFIX(auto_solve_analytical_ninf_n_intn)(ninf, n, &
         & intn, n0, intn0, t, beta, alpha, gamma, t0, emax_safety_factor, &
         & Mc, nt, use_format, validate_arrays, format_used, &
         & status) MAKECBIND
        !! Calculate the infectious aerosol concentrations for all k as time
        !! goes to infinity \(\vec{n}_\infty\) and at the provided times
        !! \(\vec{n}\) as well as its time integral at the provided times in
        !! the specified floating point format, or the smallest available
        !! floating point precision that can do the computation without
        !! overflowing if the format should be determined automatically. The
        !! recursive formulas are used. If the floating point type with the
        !! largest emax is insufficient when selecting the format
        !! automatically, the computation is not done and the situation is
        !! flagged.

        IMPLICIT NONE

        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: Mc
        !! The maximum multiplicity considered.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: nt
        !! The number of times.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(OUT) :: ninf
        !! The infectious aerosol concentration densities as time goes to
        !! infinity.
        REAL(KIND=REALP()), DIMENSION(nt, Mc), INTENT(OUT) :: n
        !! The infectious aerosol concentration densities at each time.
        REAL(KIND=REALP()), DIMENSION(nt, Mc), INTENT(OUT) :: intn
        !! The time integrals of *n* at *t*.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: n0
        !! The initial infectious aerosol concentration densities at time *t0*.
        !! Must be non-negative.
        REAL(kind=REALP()), DIMENSION(Mc), INTENT(IN) :: intn0
        !! The initial time integral of *n* values at *t0*. Must be
        !! non-negative.
        REAL(kind=REALP()), DIMENSION(nt), INTENT(IN) :: t
        !! The times to find the infectious concentration densities at. Must be
        !! greater than or equal to *t0*.
        REAL(KIND=REALP()), DIMENSION(Mc), INTENT(IN) :: beta
        !! The source vector. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: alpha
        !! The sink value. Must be non-negative.
        REAL(KIND=REALP()), INTENT(IN), VALUE :: gamma
        !! The inactivation rate. Must be non-negative.
        REAL(kind=REALP()), INTENT(IN), VALUE :: t0
        !! The initial time. Must be less than or equal to every element of
        !! *t*. Must be non-NaN and finite.
        INTEGER(KIND=C_INT), INTENT(IN), VALUE :: use_format
        !! The floating point format that should be used, or 0 to determine it
        !! automatically. See *format_used* for the values for the different
        !! floating point formats.
        REAL(kind=REALP()), INTENT(IN), VALUE :: emax_safety_factor
        !! The safety factor to add to the estimated required emax. Must be
        !! non-negative.
        INTEGER(KIND=C_INT), INTENT(OUT) :: format_used
        !! The floating point format that actually ended up being used. This is
        !! -1 if none were determined to have a large enough emax and positive
        !! otherwise, with the following meanings.
        !!
        !! Value | Fortran Kind
        !! ----- | ---------------
        !! 1     | `C_FLOAT`
        !! 2     | `C_DOUBLE`
        !! 3     | `C_LONG_DOUBLE`
        !! 4     | `REAL128`
        !!
        !! If there is an error, its value is undefined.
        LOGICAL(KIND=C_BOOL), INTENT(IN), VALUE :: validate_arrays
        !! Whether the elements of array arguments should be validated or not.
        INTEGER(KIND=C_INT), INTENT(OUT) :: status
        !! Set to 0 if successful, -1 if the arguments are invalid, -2 if
        !! memory cannot be allocated, and -3 if no available floating point
        !! format has a large enough emax.

        ! Looping index over t for validation.
        INTEGER(KIND=C_INT) :: it

        ! A variable is required to hold allocation status.
        INTEGER :: allocation_status

        ! Variables are required to hold the converted values of the arguments
        ! to be able to pass to the subroutine for the right precision.
#if defined(LIBPMADRA_CFLOAT_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_SP
        REAL(KIND=C_FLOAT), DIMENSION(:), ALLOCATABLE :: ninf_sp, n0_sp, &
             & intn0_sp, beta_sp, t_sp
        REAL(KIND=C_FLOAT), DIMENSION(:, :), ALLOCATABLE :: n_sp, intn_sp
        REAL(KIND=C_FLOAT) :: alpha_sp, gamma_sp, t0_sp
#endif
#if defined(LIBPMADRA_CDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_DP
        REAL(KIND=C_DOUBLE), DIMENSION(:), ALLOCATABLE :: ninf_dp, n0_dp, &
             & intn0_dp, beta_dp, t_dp
        REAL(KIND=C_DOUBLE), DIMENSION(:, :), ALLOCATABLE :: n_dp, intn_dp
        REAL(KIND=C_DOUBLE) :: alpha_dp, gamma_dp, t0_dp
#endif
#if defined(LIBPMADRA_CLONGDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_LD
        REAL(KIND=C_LONG_DOUBLE), DIMENSION(:), ALLOCATABLE :: ninf_ld, n0_ld, &
             & intn0_ld, beta_ld, t_ld
        REAL(KIND=C_LONG_DOUBLE), DIMENSION(:, :), ALLOCATABLE :: n_ld, intn_ld
        REAL(KIND=C_LONG_DOUBLE) :: alpha_ld, gamma_ld, t0_ld
#endif
#if defined(LIBPMADRA_REAL128_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_QP
        REAL(KIND=REAL128), DIMENSION(:), ALLOCATABLE :: ninf_qp, n0_qp, &
             & intn0_qp, beta_qp, t_qp
        REAL(KIND=REAL128), DIMENSION(:, :), ALLOCATABLE :: n_qp, intn_qp
        REAL(KIND=REAL128) :: alpha_qp, gamma_qp, t0_qp
#endif

        ! We require that Mc is positive; nt, alpha, and gamma are
        ! non-negative; and t0 is non-NaN and finite. Otherwise, there are bad
        ! arguments.
        IF (Mc <= 0 .OR. nt < 0 .OR. IEEE_IS_NAN(alpha) &
             & .OR. alpha < REALPCONST(0.) .OR. IEEE_IS_NAN(gamma) &
             & .OR. gamma < REALPCONST(0.) &
             & .OR. (.NOT. IEEE_IS_FINITE(t0))) THEN
            status = -1
            RETURN
        ENDIF

        ! Optionally validate the elements of n0, intn0, beta, and t.
        IF (validate_arrays) THEN
            IF ((.NOT. ADDSUFFIX(is_vector_nonnan_nonnegative)(n0, Mc)) &
                 & .OR. (.NOT. &
                 & ADDSUFFIX(is_vector_nonnan_nonnegative)(intn0, Mc)) &
                 & .OR. (.NOT. &
                 & ADDSUFFIX(is_vector_nonnan_nonnegative)(beta, Mc))) THEN
                status = -1
                RETURN
            ENDIF
            DO it = 1, nt
                IF (IEEE_IS_NAN(t(it)) .OR. t(it) < t0) THEN
                    status = -1
                    RETURN
                ENDIF
            END DO
        ENDIF

        ! Check use_format.
        IF (use_format < 0 .OR. use_format > 4) THEN
            status = -1
            format_used = FORMAT_VALUE_INVALID
            RETURN
        ENDIF

        ! Determine the format that should be used. If it is -2, then we need
        ! to stop.
        IF (use_format /= 0) THEN
            format_used = use_format
        ELSE
            format_used = ADDSUFFIX(required_format_solution)(n0, beta, &
                 & alpha, gamma, emax_safety_factor, Mc, .FALSE._C_BOOL)
            IF (format_used == FORMAT_VALUE_NONE) THEN
                status = -3
                RETURN
            ELSE IF (format_used < 1) THEN
                status = -1
                RETURN
            ENDIF
        ENDIF

        ! If the required format is the same as this procedure, we can just
        ! call it.
        IF (format_used == ACTUAL_FORMAT) THEN
            CALL ADDSUFFIX(solve_analytical_ninf_n_intn)(ninf, n, intn, n0, &
                 & intn0, t, beta, alpha, gamma, t0, Mc, nt, .FALSE._C_BOOL, &
                 & status)
            RETURN
        ENDIF

        ! Now, we must convert the input arguments to the right type, allocate
        ! new outputs in the right type, call the right subroutine, and convert
        ! the new outputs to the right type.

#define INCLUDE_INTEGRAL

        ! C_FLOAT
#if defined(LIBPMADRA_CFLOAT_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_SP
#define NEWSUFFIX(s) ADDSUFFIX_SP(s)
#define FORMAT_VALUE FORMAT_VALUE_SP
#define NEWREALP() C_FLOAT
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

        ! C_DOUBLE
#if defined(LIBPMADRA_CDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_DP
#define NEWSUFFIX(s) ADDSUFFIX_DP(s)
#define FORMAT_VALUE FORMAT_VALUE_DP
#define NEWREALP() C_DOUBLE
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

        ! C_LONG_DOUBLE
#if defined(LIBPMADRA_CLONGDOUBLE_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_LD
#define NEWSUFFIX(s) ADDSUFFIX_LD(s)
#define FORMAT_VALUE FORMAT_VALUE_LD
#define NEWREALP() C_LONG_DOUBLE
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

        ! REAL128
#if defined(LIBPMADRA_REAL128_SUPPORTED) && ACTUAL_FORMAT != FORMAT_VALUE_QP
#define NEWSUFFIX(s) ADDSUFFIX_QP(s)
#define FORMAT_VALUE FORMAT_VALUE_QP
#define NEWREALP() REAL128
#include "pmadra_const_coeff_auto_precision_solvers_prec_call.F90.inc"
#undef NEWSUFFIX
#undef FORMAT_VALUE
#undef NEWREALP
#endif

#undef INCLUDE_INTEGRAL

    END SUBROUTINE ADDSUFFIX(auto_solve_analytical_ninf_n_intn)
